/*
 * Copyright (c) 1998-2008 Caucho Technology -- all rights reserved
 *
 * This file is part of Resin(R) Open Source
 *
 * Each copy or derived work must preserve the copyright notice and this
 * notice unmodified.
 *
 * Resin Open Source is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Resin Open Source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, or any warranty
 * of NON-INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Resin Open Source; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   59 Temple Place, Suite 330
 *   Boston, MA 02111-1307  USA
 *
 * @author Scott Ferguson
 */

package com.caucho.config.type;

import com.caucho.config.*;
import com.caucho.config.attribute.*;
import com.caucho.config.program.*;
import com.caucho.config.types.RawString;
import com.caucho.loader.*;
import com.caucho.util.*;
import com.caucho.vfs.*;
import com.caucho.xml.QName;

import java.beans.*;
import java.io.*;
import java.net.URL;
import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.atomic.*;
import java.util.regex.Pattern;
import java.util.logging.Logger;

import javax.annotation.*;
import javax.el.*;
import javax.sql.*;

import org.w3c.dom.Node;

/**
 * Factory for returning type strategies.
 */
public class TypeFactory implements AddLoaderListener
{
  private static final Logger log
    = Logger.getLogger(TypeFactory.class.getName());
  private static L10N L = new L10N(TypeFactory.class);

  private static final HashMap<Class,ConfigType> _primitiveTypes
    = new HashMap<Class,ConfigType>();
  
  private static final EnvironmentLocal<TypeFactory> _localFactory
    = new EnvironmentLocal<TypeFactory>();

  private static final Object _introspectLock = new Object();

  private final EnvironmentClassLoader _loader;
  private final TypeFactory _parent;

  private final HashSet<URL> _configSet
    = new HashSet<URL>();
  
  private final HashMap<String,ConfigType> _typeMap
    = new HashMap<String,ConfigType>();

  private final HashMap<QName,ConfigType> _attrMap
    = new HashMap<QName,ConfigType>();

  private final HashMap<QName,Attribute> _listAttrMap
    = new HashMap<QName,Attribute>();

  private final HashMap<QName,Attribute> _setAttrMap
    = new HashMap<QName,Attribute>();

  private final HashMap<QName,Attribute> _envAttrMap
    = new HashMap<QName,Attribute>();

  private final HashMap<String,NamespaceConfig> _nsMap
    = new HashMap<String,NamespaceConfig>();

  private final HashSet<URL> _driverTypeSet
    = new HashSet<URL>();
  
  private final HashMap<String,HashMap<String,String>> _driverTypeMap
    = new HashMap<String,HashMap<String,String>>();

  private final AtomicBoolean _isInInit = new AtomicBoolean();
  
  private TypeFactory(ClassLoader loader)
  {
    _loader = Environment.getEnvironmentClassLoader(loader);

    if (_loader != null) {
      _parent = getFactory(_loader.getParent());

      _loader.addLoaderListener(this);
    }
    else
      _parent = null;

    init(loader);
  }

  /**
   * Returns the appropriate strategy.
   */
  public static ConfigType getType(Class type)
  {
    TypeFactory factory = getFactory(type.getClassLoader());

    return factory.getConfigTypeImpl(type);
  }

  /**
   * Returns the appropriate strategy.
   */
  public static ConfigType getType(Type type)
  {
    return getType((Class) type);
  }

  public static TypeFactory create()
  {
    return getFactory();
  }

  public static TypeFactory getFactory()
  {
    return getFactory(Thread.currentThread().getContextClassLoader());
  }

  public static TypeFactory getFactory(ClassLoader loader)
  {
    if (loader == null)
      loader = ClassLoader.getSystemClassLoader();

    TypeFactory factory = _localFactory.getLevel(loader);

    if (factory == null) {
      factory = new TypeFactory(loader);
      _localFactory.set(factory, loader);
      factory.init(loader);
    }

    return factory;
  }

  /**
   * Returns an environment type.
   */
  public ConfigType getEnvironmentType(QName name)
  {
    synchronized (_attrMap) {
      ConfigType type = _attrMap.get(name);

      if (type != null)
	return type;

      type = getEnvironmentTypeRec(name);

      if (type != null) {
	_attrMap.put(name, type);
	
	return type;
      }

      if (! "".equals(name.getNamespaceURI())) {
	type = getEnvironmentType(new QName(name.getLocalName()));

	if (type != null) {
	  _attrMap.put(name, type);
	
	  return type;
	}
      }

      return null;
    }
  }

  /**
   * Returns an environment type.
   */
  public ConfigType getEnvironmentTypeRec(QName name)
  {
    synchronized (_attrMap) {
      ConfigType type = _attrMap.get(name);

      if (type != null)
	return type;

      if (_parent != null)
	type = _parent.getEnvironmentTypeRec(name);

      if (type != null) {
	_attrMap.put(name, type);
	
	return type;
      }

      NamespaceConfig ns = _nsMap.get(name.getNamespaceURI());

      if (ns == null) {
	// XXX: xbeans and spring would go here
      }

      if (ns != null) {
	ns.loadBeans();

	type = ns.getBean(name.getLocalName());
	
	if (type != null) {
	  _attrMap.put(name, type);
	
	  return type;
	}
      }

      return null;
    }
  }

  /**
   * Returns an environment type.
   */
  public Attribute getListAttribute(QName name)
  {
    synchronized (_listAttrMap) {
      Attribute attr = _listAttrMap.get(name);

      if (attr != null)
	return attr;

      ConfigType type = getEnvironmentType(name);

      if (type == null)
	return null;

      attr = new ListValueAttribute(type);

      _listAttrMap.put(name, attr);

      return attr;
    }
  }

  /**
   * Returns an environment type.
   */
  public Attribute getSetAttribute(QName name)
  {
    synchronized (_setAttrMap) {
      Attribute attr = _setAttrMap.get(name);

      if (attr != null)
	return attr;

      ConfigType type = getEnvironmentType(name);

      if (type == null)
	return null;

      attr = new SetValueAttribute(type);

      _setAttrMap.put(name, attr);

      return attr;
    }
  }

  /**
   * Returns an environment type.
   */
  public Attribute getEnvironmentAttribute(QName name)
  {
    synchronized (_envAttrMap) {
      Attribute attr = _envAttrMap.get(name);

      if (attr != null)
	return attr;

      ConfigType type = getEnvironmentType(name);

      if (type == null)
	return null;

      attr = new EnvironmentAttribute(type);

      _envAttrMap.put(name, attr);

      return attr;
    }
  }

  private ConfigType getConfigTypeImpl(Class type)
  {
    synchronized (_introspectLock) {
      ConfigType strategy = _typeMap.get(type.getName());

      if (strategy == null) {
	strategy = _primitiveTypes.get(type);

	if (strategy == null)
	  strategy = createType(type);

	_typeMap.put(type.getName(), strategy);

	strategy.introspect();
      }

      return strategy;
    }
  }

  private ConfigType createType(Class type)
  {
    PropertyEditor editor = null;

    if (ConfigType.class.isAssignableFrom(type)) {
      try {
	return (ConfigType) type.newInstance();
      } catch (Exception e) {
	throw ConfigException.create(e);
      }
    }
    else if ((editor = PropertyEditorManager.findEditor(type)) != null)
      return new PropertyEditorType(type, editor);
    else if (type.getEnumConstants() != null)
      return new EnumType(type);
    else if (Set.class.isAssignableFrom(type))
      return new SetType(type);
    else if (Collection.class.isAssignableFrom(type)
	     && ! Queue.class.isAssignableFrom(type)) {
      // jms/2300
      return new ListType(type);
    }
    else if (Map.class.isAssignableFrom(type))
      return new MapType(type);
    else if (EnvironmentBean.class.isAssignableFrom(type))
      return new EnvironmentBeanType(type);
    else if (type.isArray()) {
      Class compType = type.getComponentType();
      
      return new ArrayType(getType(compType), compType);
    }
    else if (type.isInterface()
	     || Modifier.isAbstract(type.getModifiers())) {
      return new InterfaceType(type);
    }
    else
      return new BeanType(type);
  }

  /**
   * Initialize the type strategy factory with files in META-INF/caucho
   *
   * @param loader the owning class loader
   * @throws Exception
   */
  private void init(ClassLoader loader)
  {
    if (! _isInInit.getAndSet(true))
      return;

    try {
      _nsMap.clear();
      _driverTypeSet.clear();
      _driverTypeMap.clear();
      
      Enumeration<URL> urls
	= loader.getResources("META-INF/caucho/com.caucho.config.namespace.xml");

      while (urls.hasMoreElements()) {
	URL url = urls.nextElement();

	if (hasConfig(url))
	  continue;

	_configSet.add(url);

	InputStream is = Vfs.openRead(url.toString());

	try {
	  new Config(loader).configure(this, is);
	} finally {
	  is.close();
	}
      }
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw ConfigException.create(e);
    } finally {
      _isInInit.set(false);
    }
  }

  protected boolean hasConfig(URL url)
  {
    if (_configSet.contains(url))
      return true;
    else if (_parent != null)
      return _parent.hasConfig(url);
    else
      return false;
  }

  /**
   * Returns a driver by the url
   */
  public Class getDriverClassByUrl(Class api, String url)
  {
    String scheme;

    int p = url.indexOf(':');
    if (p >= 0)
      scheme = url.substring(0, p);
    else
      scheme = url;

    String typeName = getDriverType(api.getName(), scheme);

    if (typeName == null) {
      ArrayList<String> schemes = new ArrayList<String>();

      getDriverSchemes(schemes, api.getName());

      Collections.sort(schemes);
      
      throw new ConfigException(L.l("'{0}' is an unknown scheme for driver '{1}'.  The available schemes are '{2}'",
				    scheme, api.getName(), schemes));
    }

    try {
      ClassLoader loader = Thread.currentThread().getContextClassLoader();
      Class cl = Class.forName(typeName, false, loader);

      if (! api.isAssignableFrom(cl))
	throw new ConfigException(L.l("'{0}' is not assignable to '{1}' for scheme '{2}'",
				      cl.getName(), api.getName(),
				      scheme));

      return cl;
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new ConfigException(L.l("'{0}' is an undefined class for scheme '{1}'",
				    typeName, scheme), e);
    }
  }

  public ContainerProgram getUrlProgram(String url)
  {
    String properties = "";

    int p = url.indexOf(':');
    if (p >= 0) {
      properties = url.substring(p + 1);
    }
    else
      return null;

    String []props = properties.split("[;]");

    if (props.length == 0)
      return null;

    ContainerProgram program = new ContainerProgram();
    for (String prop : props) {
      if (prop.length() == 0)
	continue;
      
      String []values = prop.split("[=]");

      if (values.length != 2)
	throw new ConfigException(L.l("'{0}' is an invalid URL.  Bean URL syntax is 'scheme:prop1=value1;prop2=value2'", url));

      program.addProgram(new PropertyStringProgram(values[0], values[1]));
    }

    return program;
  }

  /**
   * Returns the classname of the given driver.
   *
   * @param apiType the driver API
   * @param scheme the configuration scheme
   */
  public String getDriverType(String apiType, String scheme)
  {
    HashMap<String,String> driverMap = getDriverTypeMap(apiType);
    
    return driverMap.get(scheme);
  }

  /**
   * Returns a list of schemes supported by the api type.
   *
   * @param schemes the return list of schemes scheme
   * @param apiType the driver API
   */
  public void getDriverSchemes(ArrayList<String> schemes, String apiType)
  {
    HashMap<String,String> driverMap = getDriverTypeMap(apiType);

    ClassLoader loader = _loader;
    if (_loader == null)
      loader = ClassLoader.getSystemClassLoader();
    
    for (Map.Entry<String,String> entry : driverMap.entrySet()) {
      String scheme = entry.getKey();
      String type = entry.getValue();

      try {
	Class cl = Class.forName(type, false, loader);

	if (cl != null)
	  schemes.add(scheme);
      } catch (Exception e) {
	log.finest(apiType + " schemes: " + e.toString());
      }
    }
  }

  /**
   * Loads the map for a driver.
   */
  private HashMap<String,String> getDriverTypeMap(String apiType)
  {
    synchronized (_driverTypeMap) {
      HashMap<String,String> driverMap = _driverTypeMap.get(apiType);

      if (driverMap == null) {
	driverMap = new HashMap<String,String>();

	if (_parent != null)
	  driverMap.putAll(_parent.getDriverTypeMap(apiType));

	loadDriverTypeMap(driverMap, apiType);

	_driverTypeMap.put(apiType, driverMap);
      }

      return driverMap;
    }
  }

  /**
   * Reads the drivers from the META-INF/caucho
   */
  private void loadDriverTypeMap(HashMap<String,String> driverMap,
				 String apiType)
  {
    try {
      ClassLoader loader = _loader;

      if (loader == null)
	loader = ClassLoader.getSystemClassLoader();
      
      Enumeration<URL> urls
	= loader.getResources("META-INF/caucho/com.caucho.config.uri/"
			      + apiType);

      while (urls.hasMoreElements()) {
	URL url = urls.nextElement();

	if (hasDriver(url))
	  continue;

	_driverTypeSet.add(url);

	InputStream is = url.openStream();

	try {
	  Properties props = new Properties();

	  props.load(is);

	  for (Map.Entry entry : props.entrySet()) {
	    driverMap.put((String) entry.getKey(), (String) entry.getValue());
	  }
	} finally {
	  is.close();
	}
      }
    } catch (Exception e) {
      throw ConfigException.create(e);
    }
  }

  protected boolean hasDriver(URL url)
  {
    synchronized (_driverTypeSet) {
      if (_driverTypeSet.contains(url))
	return true;
      else if (_parent != null)
	return _parent.hasDriver(url);
      else
	return false;
    }
  }

  //
  // AddLoaderListener
  //
  
  /**
   * Called with the loader config changes.
   */
  public void addLoader(EnvironmentClassLoader loader)
  {
    init(loader);
  }

  //
  // Configuration methods
  //

  /**
   * Adds an new environment attribute.
   */
  public NamespaceConfig createNamespace()
  {
    return new NamespaceConfig();
  }

  /**
   * Adds an new environment attribute.
   */
  public void addNamespace(NamespaceConfig ns)
  {
    _nsMap.put(ns.getName(), ns);

    if (ns.isDefault())
      _nsMap.put(ns.getName(), ns);
  }

  // configuration types
  public class NamespaceConfig {
    private String _ns = "";
    private boolean _isDefault;
    private Path _path;
    
    private AtomicBoolean _isBeansLoaded = new AtomicBoolean();

    private HashMap<String,BeanConfig> _beanMap
      = new HashMap<String,BeanConfig>();

    public void setName(String ns)
    {
      if ("default".equals(ns))
	ns = "";
      
      _ns = ns;
    }

    public String getName()
    {
      return _ns;
    }

    public void setDefault(boolean isDefault)
    {
      _isDefault = isDefault;
    }

    public boolean isDefault()
    {
      return _isDefault;
    }

    public void setPath(String path)
    {
      if (path.indexOf(':') < 0)
	_path = Vfs.lookup("classpath:" + path);
      else
	_path = Vfs.lookup(path);
    }

    public Path getPath()
    {
      return _path;
    }

    public void loadBeans()
    {
      if (_isBeansLoaded.getAndSet(true))
	return;

      new Config().configure(this, _path);
    }

    public ConfigType getBean(String name)
    {
      BeanConfig beanConfig = _beanMap.get(name);

      if (beanConfig != null)
	return beanConfig.getConfigType();
      else
	return null;
    }

    public BeanConfig createBean()
    {
      return new BeanConfig(_ns, _isDefault);
    }

    public void addBean(BeanConfig bean)
    {
      _beanMap.put(bean.getName(), bean);
    }
  }

  public class BeanConfig {
    private String _ns;
    private boolean _isDefault;

    private String _name;
    private String _className;

    private ConfigType _configType;
    private ClassLoader _loader;

    BeanConfig(String ns, boolean isDefault)
    {
      _ns = ns;
      _isDefault = isDefault;
      _loader = Thread.currentThread().getContextClassLoader();
    }

    public void setName(String name)
    {
      _name = name;
    }

    public String getName()
    {
      return _name;
    }

    public void setClass(String className)
    {
      _className = className;
    }

    public ConfigType getConfigType()
    {
      try {
	if (_configType == null) {
	  QName qName = new QName(null, _name, _ns);
	
	  Class cl = Class.forName(_className, false, _loader);

	  ConfigType type = createType(cl);

	  type.introspect();

	  _configType = type;
	}
      
	return _configType;
      } catch (Exception e) {
	throw ConfigException.create(e);
      }
    }

    @PostConstruct
    public void init()
    {
      if (_name == null)
	throw new ConfigException(L.l("bean requires a 'name' attribute"));
      
      if (_className == null)
	throw new ConfigException(L.l("bean requires a 'class' attribute"));
    }
  }

  public String toString()
  {
    return getClass().getSimpleName() + "[" + _loader + "]";
  }

  static {
    _primitiveTypes.put(boolean.class, BooleanPrimitiveType.TYPE);
    _primitiveTypes.put(byte.class, BytePrimitiveType.TYPE);
    _primitiveTypes.put(short.class, ShortPrimitiveType.TYPE);
    _primitiveTypes.put(int.class, IntegerPrimitiveType.TYPE);
    _primitiveTypes.put(long.class, LongPrimitiveType.TYPE);
    _primitiveTypes.put(float.class, FloatPrimitiveType.TYPE);
    _primitiveTypes.put(double.class, DoublePrimitiveType.TYPE);
    _primitiveTypes.put(char.class, CharacterPrimitiveType.TYPE);

    _primitiveTypes.put(Boolean.class, BooleanType.TYPE);
    _primitiveTypes.put(Byte.class, ByteType.TYPE);
    _primitiveTypes.put(Short.class, ShortType.TYPE);
    _primitiveTypes.put(Integer.class, IntegerType.TYPE);
    _primitiveTypes.put(Long.class, LongType.TYPE);
    _primitiveTypes.put(Float.class, FloatType.TYPE);
    _primitiveTypes.put(Double.class, DoubleType.TYPE);
    _primitiveTypes.put(Character.class, CharacterType.TYPE);

    _primitiveTypes.put(Object.class, ObjectType.TYPE);
    
    _primitiveTypes.put(String.class, StringType.TYPE);
    _primitiveTypes.put(RawString.class, RawStringType.TYPE);
    
    _primitiveTypes.put(String[].class, StringArrayType.TYPE);
    
    _primitiveTypes.put(Class.class, ClassType.TYPE);
    _primitiveTypes.put(Path.class, PathType.TYPE);
    _primitiveTypes.put(File.class, FileType.TYPE);
    _primitiveTypes.put(URL.class, UrlType.TYPE);
    _primitiveTypes.put(Pattern.class, PatternType.TYPE);
    _primitiveTypes.put(Locale.class, LocaleType.TYPE);
    _primitiveTypes.put(Node.class, NodeType.TYPE);
    _primitiveTypes.put(QDate.class, QDateType.TYPE);
    _primitiveTypes.put(Date.class, DateType.TYPE);
    _primitiveTypes.put(Properties.class, PropertiesType.TYPE);
    
    // _primitiveTypes.put(DataSource.class, DataSourceType.TYPE);
    
    _primitiveTypes.put(MethodExpression.class, MethodExpressionType.TYPE);
  }
}
